import functools
import typing as t

import requests
import urllib3

from . import types as tt
from .logger import LOGGER


class BasicRestClient:
    """The requests library handles mantaining a session alive and cookies for
    us.

    This class is a simple namespace to get that sorted out for us
    """

    def __init__(
        self,
        hostname: str,
        username: str,
        password: str,
        insecure: bool = False,
        port: int = 443,
    ) -> None:
        self._username = username
        self._password = password
        self._hostname = hostname
        self._insecure = insecure
        self._port = port
        self.session: requests.Session = self.prepareSession()

    def prepareSession(self) -> requests.Session:
        s = requests.Session()
        s.auth = (self._username, self._password)
        if self._insecure:
            # Disable certificate warnings
            urllib3.disable_warnings()
            s.verify = False
        self._session = s
        return self._session

    def request(
        self,
        url: str,
        params: t.Optional[t.Dict[str, t.Any]] = None,
        json: t.Optional[t.Dict[str, t.Any]] = None,
        data: t.Optional[t.Dict[str, t.Any]] = None,
        method: tt.HTTP_METHODS = "GET",
    ) -> requests.Response:
        request = functools.partial(
            self.session.request,
            method,
            f"https://{self._hostname}:{self._port}{url}",
            params=params,
            json=json,
            data=data,
            verify=not self._insecure,
        )
        try:
            res = request()
            LOGGER.debug(f"Request response: {res.text}")
            return res
        except ConnectionResetError:
            """
            If we're making too many requests the server might reset the
            connection. If so we try again but don't catch any further
            Exceptions, if they appear that means we have problems
            """
            self.prepareSession()
            return request()



import typing as t
from ipaddress import IPv4Address, IPv6Address

# Representing JSON is notoriously tricky in mypy
# Here's the best attempt I have so far
# A JSON_RES is either a list of JSON_DICTS, or just a straight up JSON_DICT
# JSON_LIST contains JSON_DICTS
# And a JSON_DICT is a simple map to acceptable JSON_VALUES
# So JSONS that contain JSONS are not acceptable, because MYPY can't represent
# self-referential values
# Meaning that if we get some sort of fancy value, we have to cast it
# to the appropriately typed dict
JSON_Values = t.Union[str, int, float, bool, None]

JSON_Dict = t.Dict[str, JSON_Values]

JSON_List = t.List[JSON_Dict]

JSON_Res = t.Any

HTTP_METHODS = t.Literal["GET", "POST", "PUT"]
STATUS = t.Literal["USED", "UNUSED"]
T = t.TypeVar("T")


def int_cast(i: t.Any) -> int:
    return int(t.cast(t.SupportsInt, i))


class Vlan(t.TypedDict, total=False):
    number: str
    name: str
    description: str
    notes: str
    vlan_id: str


class SubnetBase(t.TypedDict):
    network: str
    mask_bits: str
    name: str


class Subnet(SubnetBase, total=False):
    description: str
    notes: str


class IPAddressBase(t.TypedDict):
    """
    Only real attribute we need is a valid ipaddress
    The request method then handles converting it into an
    str
    """

    ipaddress: t.Union[IPv4Address, IPv6Address]


class IPAddress(IPAddressBase, total=False):
    label: str
    subnet: str
    macaddress: str
    device: str
    type: t.Literal["static", "dhcp", "reserved"]
    vrf_group_id: str
    vrf_group: str
    available: t.Literal["yes", "no"]
    clear_all: t.Literal["yes", "no"]
    tags: str


class StorageServiceInstance(t.TypedDict):
    service_name: t.Literal["storage_service"]
    device_id: int


class AppComponentBase(t.TypedDict):
    name: str


class AppComponent(AppComponentBase, total=False):
    device: str
    group_owner: str
    # According to the manual:
    # Description of business impact due to loss of component.
    what: str
    depends_on: str
    # Comma separated list
    dependents: str
    device_reason: str
    # list of string pairs for dependent appcomps on this appcomp e.g.
    # depend_appcomp_name1:reason1, depend_appcomp_name2:reason2
    depends_on_reasons: str


class CustomFieldBase(t.TypedDict):
    """
    Editing a custom field should be as simple as sending these to
    the relevant API.

    Getting them is a little trickier, for now I created a DOQL Query.
    """

    # ID of whichever other object you're editing
    id: int
    key: str
    value: str


class ServiceInstanceCustomField(CustomFieldBase, total=False):
    """POST/PUT: /api/1.0/custom_fields/serviceinstance

    GET: /data/v1.0/query/?saved_query_name
    =get_service_instance_custom_fields
    &delimiter=,&header=yes&output_type=json
    """

    serviceinstance_fk: int
    service_name: str
    type_id: int
    type: str
    related_model_name: t.Optional[int]
    filterable: bool
    mandatory: bool
    log_for_api: bool
    is_multi: bool
    notes: str



class Device42Exception(Exception):
    pass


class ReturnCodeException(Device42Exception):
    pass


class Device42LicenseException(Device42Exception):
    pass


class LicenseExpiredException(Device42LicenseException):
    pass


class LicenseInsufficientException(Device42LicenseException):
    pass



import argparse

parser = argparse.ArgumentParser()
parser.add_argument("-u", "--user", help="Device 42 username")
parser.add_argument("-p", "--password", help="Device 42 password")
parser.add_argument(
    "-w", "--host", help="Device 42 host. ie: https://www.(etc)"
)

args = parser.parse_args()
# Do whatever it is you want here



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pydevice42']

package_data = \
{'': ['*']}

install_requires = \
['requests>=2.25.0,<3.0.0']

setup_kwargs = {
    'name': 'pydevice42',
    'version': '0.4.1',
    'description': 'Python library to interact with Device42',
    'long_description': '[![CI](https://github.com/post-luxembourg/pydevice42/workflows/CI/badge.svg)](https://github.com/post-luxembourg/pydevice42/actions?query=workflow%3ACI)\n[![PyPI](https://img.shields.io/pypi/v/pydevice42)](https://pypi.org/project/pydevice42/)\n\n# pydevice42\n\n## Installation\n\n```shell\npip install pydevice42\n```\n\n## Usage\n\n```python\nfrom pydevice42 import D42Client\n\nclient = D42Client(hostname="device42.com", username="admin", password="admin")\n```\n',
    'author': 'Joaquim Esteves',
    'author_email': 'joaquimbve@hotmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/post-luxembourg/pydevice42',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import json as js
import typing as t
from itertools import chain

import requests

from . import exceptions as d42exc
from . import types as tt
from .basicrestclient import BasicRestClient
from .logger import LOGGER


def extract_data(data: t.Dict[str, t.Any]) -> t.Any:
    """
    When using Device42's pagination functions the return json always
    looks like so:

    ```json
    {
        "limit": INT,
        "offset": INT,
        "thing_we_care_about": (...),
        "total_count": INT
    }
    ```

    This function simply returns the first non-meta field.
    """
    metadata_keys = ["offset", "total_count", "limit"]
    return data.get(next((k for k in data.keys() if k not in metadata_keys)))


class D42Client(BasicRestClient):
    def _check_err(self, jres: t.Any) -> tt.JSON_Res:
        """POST and PUT method validation

        Raises exception if the return code isn't 0.

        Else, returns the message from the server.
        This is _generally_ a t.List[t.Any], but I have generalised it to
        be any type of tt.JSON_Res
        """
        ret_code = int(jres["code"])
        ret_msg = jres.get("msg", [])
        if ret_code != 0:
            raise d42exc.ReturnCodeException(" ".join(map(str, ret_msg)))
        return ret_msg

    def _request(
        self,
        endpoint: str,
        params: t.Optional[t.Dict[str, t.Any]] = None,
        json: t.Optional[t.Dict[str, t.Any]] = None,
        data: t.Optional[t.Dict[str, t.Any]] = None,
        method: tt.HTTP_METHODS = "GET",
    ) -> tt.JSON_Res:
        res = self.request(
            url=endpoint, params=params, json=json, data=data, method=method
        )
        try:
            res.raise_for_status()
        except requests.HTTPError as err:
            if err.response.status_code == 500:
                try:
                    msg = err.response.json().get("msg", "")
                    if msg.startswith("License expired"):
                        raise d42exc.LicenseExpiredException(msg) from err
                    elif msg.startswith("License is not valid for"):
                        raise d42exc.LicenseInsufficientException(msg) from err
                except js.JSONDecodeError:
                    # Ignore JSON decode exception here. The backend may not
                    # talk JSON when returning 500's.
                    pass
            raise
        jres: tt.JSON_Res = res.json()
        if method in ["POST", "PUT"]:
            return self._check_err(jres)
        return jres

    def _paginated_request(
        self,
        endpoint: str,
        # FIXME Is there any paginated *non-* GET request?
        method: tt.HTTP_METHODS = "GET",
        params: t.Optional[t.Dict[str, t.Any]] = None,
        json: t.Optional[t.Dict[str, t.Any]] = None,
        data: t.Optional[t.Dict[str, t.Any]] = None,
        limit: int = 4,
    ) -> t.Iterable[tt.JSON_Res]:
        def page_request(new_params: t.Dict[str, t.Any]) -> tt.JSON_Dict:
            return t.cast(
                tt.JSON_Dict,
                self._request(
                    method=method,
                    endpoint=endpoint,
                    params=new_params,
                    data=data,
                    json=json,
                ),
            )

        request_num = 1

        params = {} if params is None else params
        params["limit"] = limit
        params["offset"] = 0

        # First request
        resp = page_request(params)

        # Process data
        total_count = tt.int_cast(resp.get("total_count"))
        resp_data: t.Any = extract_data(resp)
        yield resp_data

        processed = len(resp_data)

        while processed < total_count:
            params["offset"] += limit
            request_num += 1
            LOGGER.debug(
                f"Processing request #{request_num}) "
                f"[Offset: {params['offset']} - Limit: {limit}] "
                f"{len(resp_data)}/{total_count}"
            )
            resp = page_request(params)
            resp_data = extract_data(resp)
            processed += len(resp_data)
            yield resp_data

    def _flattened_paginated_request(
        self, *args: t.Any, **kwargs: t.Any
    ) -> t.Iterable[tt.JSON_Res]:
        return chain.from_iterable(
            t.cast(
                t.List[tt.JSON_Res], self._paginated_request(*args, **kwargs)
            )
        )

    def post_network(self, new_subnet: tt.Subnet) -> tt.JSON_Res:
        return self._request(
            endpoint="/api/1.0/subnets/",
            method="POST",
            data=t.cast(t.Dict[str, t.Any], new_subnet),
        )

    def post_ip(self, new_ip: tt.IPAddress) -> tt.JSON_Res:
        return self._request(
            endpoint="/api/1.0/ips/",
            method="POST",
            data=t.cast(t.Dict[str, t.Any], new_ip),
        )

    def post_app_component(self, new_component: tt.AppComponent) -> tt.JSON_Res:
        return self._request(
            endpoint="/api/1.0/appcomps/",
            method="POST",
            data=t.cast(t.Dict[str, t.Any], new_component),
        )

    def get_DOQL_query(self, query_name: str) -> t.Any:
        """
        DOQL queries are custom usermade queries that talk directly to
        the database and (generally) return a JSON.

        They have to be coded by the user.
        """
        return self._request(
            method="GET",
            endpoint="/services/data/v1.0/query/",
            params={
                "saved_query_name": query_name,
                "delimiter": "",
                "header": "yes",
                "output_type": "json",
            },
        )

    def update_custom_field(
        self, cf: tt.CustomFieldBase, endpoint: str
    ) -> tt.JSON_Res:
        """
        Update a custom field for a given d42 object.

        Note that a CustomFieldBase's `value` is a string!

        This means that if you're inputting a json as a custom field
        you should cast it with `json.dumps`

        Example:

        ```python
        >>> client = D42Client(user, password, host)
        >>> client.update_custom_field(
        ...             {
        ...                 "id": 12,
        ...                 "key": "custom_data",
        ...                 "value": {
        ...                     "Testing": "I was sent from the API"
        ...                 },
        ...             },
        ...             "serviceinstance"
        ...         )
        Traceback (most recent call last):
        ...
        requests.exceptions.RequestException: <Response [500]>
        >>> client.update_custom_field(
        ...             {
        ...                 "id": 12,
        ...                 "key": "custom_data",
        ...                 "value": dumps({
        ...                     "Testing": "I was sent from the API"
        ...                 }),
        ...             },
        ...             "serviceinstance"
        ...         )
        (0, 'custom key pair values added or updated ...')
        ```
        """

        return self._request(
            method="PUT",
            endpoint=f"/api/1.0/custom_fields/{endpoint}/",
            data=t.cast(t.Dict[str, t.Any], cf),
        )

    def get_all_devices(self) -> tt.JSON_Res:
        return self._request(
            method="GET",
            endpoint="/api/1.0/devices/all/",
        ).get("Devices")

    def get_all_service_instances(self) -> tt.JSON_Res:
        return self._flattened_paginated_request("/api/2.0/service_instances/")

    def get_all_application_components(self) -> tt.JSON_Res:
        return self._flattened_paginated_request("/api/2.0/appcomps/")

    def get_all_operating_systems(self) -> tt.JSON_Res:
        return self._flattened_paginated_request("/api/1.0/operatingsystems/")
